#include "io.h"
#include "mqtt.h"
#include "util.h"

#include <algorithm>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"

#include "esp_log.h"
#include "driver/gpio.h"

#define IO_TAG "IO"

#define ESP_INTR_FLAG_DEFAULT 0
#define STATUS_LED_PIN GPIO_NUM_2

#define DEBOUNCE_MILLIS 10
const TickType_t debounce_delay_ticks = std::max(DEBOUNCE_MILLIS / portTICK_PERIOD_MS, (TickType_t) 1);

struct button_t {
    TickType_t last_activation_millis = 0;
    gpio_num_t gpio_num;
    // gpio starts with pullup enabled
    bool last_level = 1;
    std::function<void()> on_press;
    std::function<void()> on_release;
};

bool interrupts_enabled = false;
xQueueHandle gpio_event_queue = NULL;

void check_button_level(button_t* button) {
    bool level = gpio_get_level(button->gpio_num);

    if (level == button->last_level) {
        return;
    }

    ESP_LOGI(IO_TAG, "GPIO level changed: gpio %d, level %d", button->gpio_num, level);
    if (level) {
        if (button->on_release != NULL) {
            button->on_release();
        }
    }
    else {
        if (button->on_press != NULL) {
            button->on_press();
        }
    }

    button->last_level = level;
}

void interrupt_queue_task(void* arg) {
    button_t* button;
    while (true) {
        if (xQueueReceive(gpio_event_queue, &button, portMAX_DELAY)) {
            // immediately react to changing edge
            // value might be incorrect due to contact bouncing, but that's ok, it is read again after debouncing
            check_button_level(button);

            // wait for debouncing to finish
            bool wait;
            do {
                vTaskDelay(debounce_delay_ticks);
                TickType_t now = xTaskGetTickCount() * portTICK_PERIOD_MS;
                wait = now < button->last_activation_millis + DEBOUNCE_MILLIS;
            } while(wait);

            // handle button again after debouncing
            check_button_level(button);
        }
    }
}

void enable_interrupts() {
    if (!interrupts_enabled) {
        ESP_ERROR_CHECK(gpio_install_isr_service(ESP_INTR_FLAG_DEFAULT));
        gpio_event_queue = xQueueCreate(10, sizeof(uint32_t));
        xTaskCreate(interrupt_queue_task, "interrupt_queue_task", 2048, NULL, 10, NULL);

        interrupts_enabled = true;
        ESP_LOGI(IO_TAG, "Interrupts enabled");

        ESP_LOGI(IO_TAG, "portTICK_PERIOD_MS = %d", portTICK_PERIOD_MS);
    }
}

void IRAM_ATTR gpio_isr_handler(void* arg) {
    button_t* button = (button_t*) arg;
    TickType_t now = xTaskGetTickCountFromISR() * portTICK_PERIOD_MS;
    if (now >= button->last_activation_millis + DEBOUNCE_MILLIS) {
        xQueueSendFromISR(gpio_event_queue, &button, NULL);
    }
    else {
        button->last_activation_millis = now;
    }
    button->last_activation_millis = now;
}

void add_button(gpio_num_t gpio_num, std::function<void()> on_press, std::function<void()> on_release) {
    enable_interrupts();

    button_t* button = new button_t;
    button->gpio_num = gpio_num;
    button->on_press = on_press;
    button->on_release = on_release;

    gpio_pad_select_gpio(gpio_num);
    ESP_ERROR_CHECK(gpio_set_direction(gpio_num, GPIO_MODE_INPUT));
    ESP_ERROR_CHECK(gpio_pullup_en(gpio_num));
    ESP_ERROR_CHECK(gpio_set_intr_type(gpio_num, GPIO_INTR_ANYEDGE));

    ESP_ERROR_CHECK(gpio_isr_handler_add(gpio_num, gpio_isr_handler, button));
}

void add_button(gpio_num_t gpio, std::string topic, std::string message) {
    add_button(gpio, [topic, message](){ publish_message(topic, message); });
}

void set_status_led(uint32_t level) {
    ESP_ERROR_CHECK(gpio_set_level(STATUS_LED_PIN, level));
}

void gpio_init_output(gpio_num_t gpio_num, uint32_t level) {
    gpio_pad_select_gpio(gpio_num);
    ESP_ERROR_CHECK(gpio_set_direction(gpio_num, GPIO_MODE_OUTPUT));
    ESP_ERROR_CHECK(gpio_set_level(gpio_num, level));
}

void pulse_relay(gpio_num_t gpio) {
    ESP_ERROR_CHECK(gpio_set_level(gpio, 1));
    vTaskDelay(25 / portTICK_PERIOD_MS);
    ESP_ERROR_CHECK(gpio_set_level(gpio, 0));
}

void add_relay(const std::string& topic, gpio_num_t gpio_off, gpio_num_t gpio_on) {
    gpio_init_output(gpio_off, 0);
    gpio_init_output(gpio_on, 0);

    add_message_callback(topic, [gpio_off, gpio_on](std::string message){
        if (message == "0") {
            pulse_relay(gpio_off);
        }
        else if (message == "1") {
            pulse_relay(gpio_on);
        }
    });
}
